/* rc5pipe by HardCore Software
   (C) 2002 HardCore Software

   Distributed under the terms and conditions of the GNU GPL.
   See the file "LICENSE" for details.

   Some code taken from RSA RC5 reference implementation.
   See copyright below.
*/

/* RC5REF.C -- Reference implementation of RC5-32/12/16 in C.        */
/* Copyright (C) 1995 RSA Data Security, Inc.                        */

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#define VERSION "1.1"
#define MAXLINESIZE 2048
#define PADDINGCEILING 10 /* HIGHER THAN 10 AND THIS WILL BREAK! */

typedef unsigned long int WORD; /* Should be 32-bit = 4 bytes        */
#define w        32             /* word size in bits                 */
#define r        12             /* number of rounds                  */  
#define b        16             /* number of bytes in key            */
#define c         4             /* number  words in key = ceil(8*b/w)*/
#define t        26             /* size of table S = 2*(r+1) words   */
WORD S[t];                      /* expanded key table                */
WORD P = 0xb7e15163, Q = 0x9e3779b9;  /* magic constants             */

/* Rotation operators. x must be unsigned, to get logical right shift*/
#define ROTL(x,y) (((x)<<(y&(w-1))) | ((x)>>(w-(y&(w-1)))))
#define ROTR(x,y) (((x)>>(y&(w-1))) | ((x)<<(w-(y&(w-1)))))

void RC5_ENCRYPT(WORD *pt, WORD *ct) /* 2 WORD input pt/output ct    */
{ WORD i, A=pt[0]+S[0], B=pt[1]+S[1];
  for (i=1; i<=r; i++) 
    { A = ROTL(A^B,B)+S[2*i]; 
      B = ROTL(B^A,A)+S[2*i+1]; 
    }
  ct[0] = A; ct[1] = B;  
} 

void RC5_DECRYPT(WORD *ct, WORD *pt) /* 2 WORD input ct/output pt    */
{ WORD i, B=ct[1], A=ct[0];
  for (i=r; i>0; i--) 
    { B = ROTR(B-S[2*i+1],A)^A; 
      A = ROTR(A-S[2*i],B)^B; 
    }
  pt[1] = B-S[1]; pt[0] = A-S[0];  
} 

void RC5_SETUP(unsigned char *K) /* secret input key K[0...b-1]      */
{  WORD i, j, k, u=w/8, A, B, L[c]; 

  if (sizeof(WORD)!=4) { 
    fprintf(stderr, "rc5pipe: FATAL ERROR! A long int should be 4 bytes, but it's %d on yours.\n",sizeof(WORD));
    exit(0);
  }

   /* Initialize L, then S, then mix key into S */
   for (i=b-1,L[c-1]=0; i!=-1; i--) L[i/u] = (L[i/u]<<8)+K[i];
   for (S[0]=P,i=1; i<t; i++) S[i] = S[i-1]+Q;
   for (A=B=i=j=k=0; k<3*t; k++,i=(i+1)%t,j=(j+1)%c)   /* 3*t > 3*c */
     { A = S[i] = ROTL(S[i]+(A+B),3);  
       B = L[j] = ROTL(L[j]+(A+B),(A+B)); 
     } 
} 


int rc5decloop() { 

   char decbuf[(MAXLINESIZE*4) + PADDINGCEILING + 10];
   char realbuf[MAXLINESIZE+10];
   char *buf;
   char tpbuf[9];
   WORD inpair[2];
   WORD outpair[2];
   int i,j;

   memset(realbuf, 0, sizeof(realbuf));
   memset(tpbuf, 0, sizeof(tpbuf));
   buf = decbuf;

   while(fgets(buf, MAXLINESIZE*4, stdin)) {

   if (strncmp(buf, "RC5-", 4) != 0) {
     fprintf(stderr, "rc5pipe: Error in decoding a line. continuing with next.\n");
     continue;
   }
   buf += 4; /* Skip "RC5-" header */

   if (strlen(buf) % 16 != 1) {
     fprintf(stderr, "rc5pipe: Skipping line of bad decoding input.\n");
     continue;
   }


   if (buf[strlen(buf)-1] == '\n') buf[strlen(buf)-1]='\0';

     for(i=0;i<strlen(buf);i+=16) {

       memset(inpair, 0, sizeof(inpair));
       memset(outpair, 0, sizeof(outpair));

       strncpy(tpbuf, buf+i, 8);
       sscanf(tpbuf, "%X", &inpair[0]);
       strncpy(tpbuf, buf+i+8, 8);
       sscanf(tpbuf, "%X", &inpair[1]);

       RC5_DECRYPT(inpair, outpair);  

       memset(tpbuf, 0, sizeof(tpbuf));
       memcpy(tpbuf, outpair, 8);
       strcat(realbuf, tpbuf);

     }

     if (realbuf[0] < 48 || realbuf[0] > 57) {
       fprintf(stderr, "rc5pipe: Either password wrong, or padding is fucked. Decrypting anyways.\n");
       printf("%s", realbuf);
     } else {
       printf("%s", realbuf+((realbuf[0]-48)+1));
       printf("\n");
     }
     fflush(stdout);

     buf = decbuf;
     memset(decbuf, 0, sizeof(decbuf));
     memset(realbuf, 0, sizeof(realbuf));
   }

  return 1;
}



int addpadding(char *buf) {

  int len,i;

  sprintf(buf, "%d", len = (rand() % PADDINGCEILING));

  for (i=0;i<len;i++) { 
    buf[i+1] = (unsigned char)(rand() % 256); 
    if (buf[i+1] == '\0') buf[i+1] = 'A';
  }

  return 1;

}



int rc5encloop() { 

   char buf[MAXLINESIZE];
   WORD inpair[2];
   WORD outpair[2];
   int i,j;

   srand(time(NULL)+getpid()); /* Seed random */

   memset(buf, 0, sizeof(buf));

   while(addpadding(buf) && fgets(buf+strlen(buf), sizeof(buf)-1, stdin)) {

   if (strlen(buf) >= sizeof(buf)-1)
     fprintf(stderr, "rc5pipe: Line too long. Wrapping the rest of the line.\n", MAXLINESIZE);


   printf("RC5-");

   if (buf[strlen(buf)-1] == '\n') buf[strlen(buf)-1]='\0';

     for(i=0;i<strlen(buf);i+=8) {

       memset(outpair, 0, sizeof(outpair));
       memset(inpair, 0, sizeof(inpair));

       memcpy(inpair, buf+i, 8);

       RC5_ENCRYPT(inpair, outpair);  

       printf("%.8lX%.8lX", outpair[0], outpair[1]);

     }

     printf("\n");

     fflush(stdout);
     memset(buf, 0, sizeof(buf));
   }

  return 1;
}


void usage() {

  fprintf(stderr, "rc5pipe V%s\n(C) 2002 by HardCore Software\n\n", VERSION);
  fprintf(stderr, "Usage: rc5pipe -[e|d|se|sd] [RC5 password]\n");
  fprintf(stderr, "        -e Encrypt from stdin to stdout with key in following argument\n");
  fprintf(stderr, "        -d Decrypt from stdin to stdout with key in following argument\n");
  fprintf(stderr, "        -se Same as -e, but instead prompt for password\n");
  fprintf(stderr, "        -sd Same as -d, but instead prompt for password\n");
  fprintf(stderr, "\n");

  exit(0);

}




int main(int argc, char *argv[]) { 

  int mode=-1;

  if ((argc > 3) || (argc < 2) || (argv[1][0] != '-')) usage();

  if (argc == 3) {
    if (argv[1][1] == 'e') { RC5_SETUP(argv[2]); mode=0; } 
    if (argv[1][1] == 'd') { RC5_SETUP(argv[2]); mode=1; }
  } else {
    if (argv[1][1] == 's') {
      if (argv[1][2] == 'e') { RC5_SETUP(getpass("RC5 key? ")); mode=0; } 
      if (argv[1][2] == 'd') { RC5_SETUP(getpass("RC5 key? ")); mode=1; }
    }
  }

  if (mode==-1) usage();
  if (mode==0) rc5encloop();
  if (mode==1) rc5decloop ();

  return 1;

}
